/* (c) Joerg Schulenburg 2000-2019

 this is mpi/pthread-specific part of spin-package

 it is for division of matrix multiplication 
 in parallelizable block matrix multiplication
 
    b = H * a    (a,b vectors, H matrix)
    
    b[i] = sum_j H[i,j] * a[j]
    
 blocks can be multiplicated step by step on one system or
 they can be multiplicated on different processors/systems
 
 but you have to pay additional network/harddisk traffic
 
 if b_num is one you have only one block (similar to old version)
 
 - use mpi_n nodes and mpt_n (1..B_NUM) threads per node
 - variable block counts from 0 to pt_n-1 (B_NUM) in each node
 - set B_NUM or CFG_PTHREAD only if using OpenMP or pthreads
   it eats unused memory if bigger than 1 and no threading used (MPI-only)
   B_NUM can and should be set by Makefile.CFG_PTHREAD=... since 2016-05-18 
   B_NUM will be removed in future, default = CFG_PTHREAD = 1

 */

#ifndef _HBLOCK_H
#define _HBLOCK_H

#ifdef CONFIG_PTHREAD
/* 2017-02 mpiCC + icpc16 does not like it */
//#define _XOPEN_SOURCE 600  /* needed to define Barrier prototypes ??? */
#endif

#if defined(CFG_MPI) && defined(CONFIG_MPI)
#error only one MPI-switch must be defined outside /* 2016-05-18 debug */
#endif
#if defined(CFG_MPI) && !defined(CONFIG_MPI)
#define CONFIG_MPI CFG_MPI   /* 2016-05-18 shorter macro can be used too */
#endif
#if defined(CONFIG_MPI) && !defined(CFG_MPI)
#define CFG_MPI CONFIG_MPI   /* 2016-05-18 compatibility, ToDo: obsolete */
#endif

#ifdef CONFIG_MPI
#if  defined(__cplusplus) && !defined(MPICH_IGNORE_CXX_SEEK)
/* may be including stdio.h before mpi.h will work too */
#define MPICH_IGNORE_CXX_SEEK  /* avoid macro conflicts mpich with stdio.h */
#endif /* __cplusplus */
#include <mpi.h>   /* CXX_SEEK defined by mpich conflicts with stdio.h */
#endif /* CONFIG_MPI */

#include "config.h"

/* maximum number of compute threads per mpi node, set by option -t<num>
 *  else set by OMP_NUM_THREADS (2016-05)
 * check nodes for: grep -c ^processor /proc/cpuinfo
 *   or: pbsnodes|grep "^\ *np\ =" |sort -u -k 3n|head -1|awk '{print $3}'
 *   or: sinfo -h -o %c
 */
#ifndef B_NUM  /* see remaks at top */
#if     defined(CONFIG_PTHREAD) && CONFIG_PTHREAD>1
#define B_NUM CONFIG_PTHREAD /* 2016-05-18 can be set by configure/Makefile now */
#elif   defined(CONFIG_PTHREAD) && !defined(CONFIG_MPI)
#define B_NUM 32 /* may be 1000 or more on massive SMP systems */
#elif   defined(CONFIG_PTHREAD) &&  defined(CONFIG_MPI) /* 2016-04-22 ~! */
#define B_NUM 8  /* spare memory on small nodes */
#else
#define B_NUM 1  /* set to minimum to spare memory */
#endif
#endif

/* ToDo2016 autodetect MPI_MAX_NUM by:
 *   pbsnodes | grep -c ^[a-z]  # comp2
 * x pbsnodes|grep "^\ *np\ =" |sort -u -k 3n|head -1|awk '{print $3}'
 *   sinfo -h -o %D             # t100,kautz
 * x sinfo -h -o %c             #  if B_NUM=1
 * 2017: 16K tasks (1024nodes *16cores)
 * 2017-01-18 enlarge from 8K to 64K/B_NUM fits to int16 (*4B=256KB), bigger default
 *   used for Alltoallv-tables and l1_ofs (MPI-transfer-buffers)
 * 2013 lrz.supermuc.phase1.thin 9216nodes*16c/n*2t/c 16GB/n jobmax=4096*16 2d
 * 2015 lrz.supermuc.phase2      3072nodes*28c/n*2t/c 64GB/n jobmax= 512*28 2d
 * 2019 lrz.supermuc-NG          6336nodes*48c/n*2t/c 96GB/n jobmax=3168*48 2d
 *  use system-spec. env CONFIG_MPI_MAX_NUM + CONFIG_MPT_MAX_NUM in configure
 */
#ifndef MPI_MAX_NUM   /* ToDo2016: use CONFIG_MPI > 1 to set mpi_max_num */
#  ifndef CONFIG_MPI
#    define  MPI_MAX_NUM 1          /* minimum bufsize, added  2018-11-20 */
#  elif   CONFIG_MPI > 1            /* if-not-defined macro is 0 */
#    define  MPI_MAX_NUM CONFIG_MPI  /* added  2018-11-20 ToDo: use sh.ENV */
#  else
#    define  MPI_MAX_NUM (2048*(1+31/(B_NUM)))  /* dflt max num MPI tasks */
#  endif  // CONFIG_MPI
#endif  // MPI_MAX_NUM

/* #include "defs.h" */

extern int mpi_n, mpi_i;	/* number of MPI nodes + nodenr.         */
extern int pt_n;                /* number of threads per node (1..B_NUM) */
extern int pt_n_bits, mpi_n_bits; /* bits needed to store pt_i and mpi_i */
/* pointer for blocked and locally stored vectors (MPI!) */
/* see b_ini: node_ofs=mpi_ofs[mpi_i]; b_ofs[i]=0..node_n1 */
extern unsigned long node_ofs, node_n1;	/* part of space on this node */
/* pointer for blocked and locally stored vectors (NUMA!) [mpi_n*pt_n] */
// extern unsigned long *b_ofs;	/* start index 0..n1-1 of block (obsolete) */
// extern unsigned long *b_len;	/* block len (obsolete) */
// extern unsigned long max_b_len; /* maximum b_len to ensure mpi balance */

/* this offset definitions are new in v2.36, and replace old pt_n*mpi_n
 *    which have required aditional operations in the inner loop 
 * start index 0..n1 of vector in thread[0..pt_n] for this mpi node */
extern unsigned long b_ofs[B_NUM+1];
/* start index 0..n1 of vector in mpi_node[0..mpi_n] */
extern unsigned long mpi_ofs[MPI_MAX_NUM+1];
/* absolute ofs: mpi_ofs[mpi_i]+b_ofs[pthread_i]
 * block length: mpi_ofs[mpi_n]==n1
 *               b_ofs[pt_n]=node_n1
 *                          =mpi_ofs[mpi_i+1]-mpi_ofs[mpi_i]
 */
//extern unsigned long max_b_len; /* maximum block length ensure mpi-balance */

/* this is for stored files, ToDo: move to h_file */
extern unsigned long b_hnz[B_NUM];	/* NumNonZero-H block entries */
/* this is for partly stored files, ToDo: remove or replace */
extern unsigned long b_snz[B_NUM];	/* stored NonZero-H block entries */
//extern unsigned long b_sn1[B_NUM];	/* stored block lines, new _ah_blks */
extern int b_bits;			/* b_len[0..n-2]=1l<<b_bits */
extern unsigned long b_mask;		/* ~((~0)<<b_bits) */
extern unsigned long len_full_block;   /* number of elements per block */

/* initialization, n1 is whole matrix size */
int b_ini(unsigned long n1,int vvv);
void b_free(); /* opposit of b_ini */

//#include "h_file.h"  /* 2019-01: move thxy+b_hbuf to hilbert.h+thbuf */
//extern struct thxy b_hbuf[B_NUM]; /* buffer for faster access, [HBLen] */

extern int          b_bnz[B_NUM]; /* num valid nz-data in buffer bufh */  
extern int          b_bn1[B_NUM]; /* num valid lines in buffer bufh */  

/* 2015-12-31 testwise, ToDo: pthread-alternative */
#ifdef _OPENMP
#include <omp.h>
#else
#define omp_get_max_threads() 1
#define omp_get_thread_num() 0
#endif

/* 2017-02-02 moved here from spins.c */
#ifdef CONFIG_PTHREAD   /* ToDo: movto mthreads.h, or write new OpenMP-code */
#ifndef _OPENMP /* OPENMP is replacing PTHREAD */
#include <pthread.h>
/* pt_attr: SunOS has SCOPE_PROCESS as default, we need SCOPE_SYSTEM */
extern pthread_attr_t pt_attr;
extern pthread_cond_t pt_cond1;  /* conditions master -> childs */
extern pthread_cond_t pt_cond2;  /* conditions childs -> master */
extern pthread_mutex_t pt_mutex1; /* mutex */
extern pthread_mutex_t pt_mutex2; /* mutex */
/* MPI code is in not thread safe in general, we have to care about it */
#if defined(__APPLE__)   /* minimum implementation by A.Armea 2013, 2018-11 */
/* there is no pthread_barrier-implementation on macOS-10.14 */
/* ToDo18: future versions may have barrier implemented so better rename
 *         functions to my_barrier and add wrapper for ~__APPLE__ */
#ifndef PTHREAD_BARRIER_H_
#define PTHREAD_BARRIER_H_
//#include <pthread.h>
#include <errno.h>
typedef int pthread_barrierattr_t;
typedef struct {
    pthread_mutex_t mutex;
    pthread_cond_t cond;
    int count;
    int tripCount;
} pthread_barrier_t;
int pthread_barrier_init(pthread_barrier_t *barrier,
    const pthread_barrierattr_t *attr, unsigned int count);
int pthread_barrier_destroy(pthread_barrier_t *barrier);
int pthread_barrier_wait(pthread_barrier_t *barrier);
#endif // PTHREAD_BARRIER_H_
#endif // __APPLE__
extern pthread_mutex_t token_ring[B_NUM*4]; /* 4 rings of double(quad) length */
/*   grep pthread_barrier_t /usr/include/bits/pthreadtypes.h 
 * mips: pthread.h = gentoo-multilib/n64/pthread.h 2006
 *       #ifdef __USE_XOPEN2K pthread_barrier_t 
 * user should set _XOPEN_SOURCE or _POSIX_C_SOURCE or -std=gnu89|90|99
 *  gcc492 -std=c89 will fail with unknown pthread_barrier_t
 * macOS Apple-LLVM-10.0=clang has no pthread_barrier_t by default
 */
extern pthread_barrier_t pt_barrier; /* 2017-01-18 for new better mixed code (MPI+PT) */
#else /* OPENMP-replacements? PTHREAD-dummys! */
#define pthread_mutex_lock(x)
#define pthread_mutex_unlock(x)
#define pthread_barrier_wait(x) omp_barrier_wait()
extern volatile int omp_token;
extern volatile int omp_token2;
static inline int omp_barrier_wait() {
#pragma omp atomic   // omp-3.0 only? gcc412.omp2.5=OK? need flush? ToDo17
    omp_token++; // OMP_BARRIER cant be used, because of our master thread?
    while (omp_token%pt_n!=0); /* usleep? */
    omp_token2++; // OMP_BARRIER cant be used, because of our master thread?
    while (omp_token2%pt_n!=0); /* usleep? */
    return 0;
}
#define pthread_cond_broadcast(x)
#endif /* ~_OPENMP */
#else  /* ~PTHREAD dummy definitions to eleminate ifdef constructions */
#define pthread_mutex_lock(x)
#define pthread_mutex_unlock(x)
#define pthread_barrier_wait(x)
#define pthread_cond_broadcast(x)         
#endif

// Jul22 moved from spins.c to block.h
/* local variables for pthreads (tasks), communication to/from tasks */
typedef struct {
  int num;     /* to identify itself (ro) */
  void *data;  /* to send data-block (set by thread 0, read by workers) */
  int  task;   /* master set 1 to run task + waits for 0, task sets 0 @end */
  void (*worker)(void *data); /* parallel working function */
  /* mutex for each thread? */
} main_pts;
extern volatile main_pts main_pt[B_NUM]; /* [un]locked by mutex1 */

//#ifdef CONFIG_PTHREAD
/* send void-function + data to waiting parallel thread */
/* 2016-10 g++-6.2 + pthread + blas segfaults for stacksize 8K (ulimit -s) */
/*  give waiting thread i a void-function and a data-pointer to work */
/*  ... usually this is done via loop over all i to let pthread_barrier work */
void pt_start( int thread_i, void (*worker)(void *data), void *data );
/* waiting for threads finishing there task given by pt_start */
void pt_end_wait( int thread_i ); // Jul22: renamed from parallel_wait
// PT_Barrier + rc = (int)1 for one thread, rest get 0
//#endif // CONFIG_PTHREAD
int pt_barrier_wait();  // sync pthreads, needed for hybrid MPI/PT-code

#ifndef CONFIG_MPI  /* define MPI dummy-functions to reduce ifdef-lines */
#define MPI_Barrier(MPI_COMM_WORLD)  /* dummy function for ~MPI */
#define MPI_Bcast(abuf,cnt,MPI_INT,root,MPI_COMM_WORLD) /* 2012-05 */
// #define MPI_Bcast(x1,x2,x3,x4,x5) /* dummy 2012-05 */
#define MPI_Abort(MPI_COMM_WORLD,x2)  exit(x2) /* dummy*/
// MPI_Allreduce(&add, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
#define MPI_Allreduce(add, sum, cnt, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD)\
   { int i; for (i=0;i<(cnt);i++) (sum)[i]=(add)[i]; } /* 2016-04-15 */
#endif // ~CONFIG_MPI

#endif   /* ifndef _HBLOCK_H */
